/*
Copyright (c) 2013 Filippo Mariani

This software is provided 'as-is', without any express or implied warranty.
In no event will the authors be held liable for any damages arising from the use of this software.
Permission is granted to anyone to use this software for any purpose, 
including commercial applications, and to alter it and redistribute it freely, 
subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must not claim that you wrote the original software. If you use this software in a product, an acknowledgment in the product documentation would be appreciated but is not required.
2. Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.
3. This notice may not be removed or altered from any source distribution.
*/

// The best documentation available about VHACD parameters can be found here:
// http://kmamou.blogspot.com/

#ifndef BTVHACDCOMPOUNDSHAPE_H_
#define BTVHACDCOMPOUNDSHAPE_H_
#pragma once


#include <vhacdHACD.h>
#include <vhacdMeshDecimator.h>
#include <vhacdMesh.h>

#include <BulletCollision/CollisionShapes/btCompoundShape.h>
#include <BulletCollision/CollisionShapes/btConvexHullShape.h>
#include <BulletCollision/CollisionShapes/btStridingMeshInterface.h>
#include <BulletCollision/CollisionShapes/btShapeHull.h>
#include <BulletCollision/CollisionShapes/btBvhTriangleMeshShape.h>
#include <BulletCollision/CollisionShapes/btScaledBvhTriangleMeshShape.h>
#include <BulletCollision/Gimpact/btGImpactShape.h>
#include <LinearMath/btGeometryUtil.h>
#include <LinearMath/btConvexHullComputer.h>
#include <LinearMath/btSerializer.h>
#include <ctime>
#include <vector>
#include <algorithm>

#pragma region class  btVHACDCompoundShape : public btCompoundShape
// Does NOT delete child shape on exiting (user may want to reuse them for other shapes, like when making a btCompoundShape out of uniform scaled children of this shape)
ATTRIBUTE_ALIGNED16(class)  btVHACDCompoundShape : public btCompoundShape
{
public:

struct Params {
    // ORIGINAL PARAMS FROM CURRENT VHACD GIT
    size_t depth;                           // (10)     maximum number of decomposition stages
    double maxConcavity;                    // (0.01)   maximum allowed concavity
    //bool invertInputFaces;                // (false)  indicates whether mesh normals should be inverted or not
    bool displayDebugInfo;                  // (false)  displays debug info through printf
    size_t posSampling;                     // (10)     clipping plane position sampling resolution for coarse search
    size_t angleSampling;                   // (10)     clipping plane orientation sampling resolution for coarse search
    size_t posRefine;                       // (5)      clipping plane position sampling resolution for refined search
    size_t angleRefine;                     // (5)      clipping plane orientation sampling resolution for refined search
    double alpha;                           // (0.01)   parameter controlling the compromise between concavity and balance between convex-hulls
    size_t targetNTrianglesDecimatedMesh;   // (1000)   number of triangles in the decimated mesh. V-HACD decimates the input mesh to reduce computation times
    //std::string optionalVRMLSaveFilePath;   // ("")

	// W.I.P.: ADDITIONAL PARAMS ADDED BY ME (TO BE TESTED)
        #define BTHACDCOMPOUNDSHAPE_H_USES_STDVECTOR_AS_VECTOR       // To check: without this definition, I got errors in some configurtions (but in another part of the code).
            #ifdef BTHACDCOMPOUNDSHAPE_H_USES_STDVECTOR_AS_VECTOR
            typedef std::vector< int > IntVector;
            #else //BTHACDCOMPOUNDSHAPE_H_USES_STDVECTOR_AS_VECTOR
            typedef btAlignedObjectArray< int > IntVector;            
            #endif //BTHACDCOMPOUNDSHAPE_H_USES_STDVECTOR_AS_VECTOR

        IntVector decomposeOnlySelectedSubmeshes;                   // if size()==0, decompose all submeshes
        bool keepSubmeshesSeparated;									// (false). True is slower, and must be used when submesh index is needed.
																// A btAlignedObjectArray< int > submeshIndexOfChildShapes can be retrieved from the class instance to map the child shape index vs the submesh index.
        //IntVector keepSubmeshesSeparatedNClusters;				// Optional (used if size()>0 and keepSubmeshesSeparated==true). If any value is missing or <= 0, the global nClusters is used instead.
																// Ordering refers to the whole set of submeshes (so that it's independent on decomposeOnlySelectedSubmeshes).
        //IntVector keepSubmeshesSeparatedMaxHullVertices;		// Optional (used if size()>0 and keepSubmeshesSeparated==true). If any value is missing or < 0, the global maxHullVertices is used instead.
																// Ordering refers to the whole set of submeshes (so that it's independent on decomposeOnlySelectedSubmeshes).														
	bool decomposeACleanCopyOfTheMesh;							// (true). Handles duplicated vertices and degenerate triangles. Slower.
	bool decomposeADecimatedCopyOfTheMesh;						// (false). (DEPRECATED: Use targetNTrianglesDecimatedMesh instead). True can be used to reduce the number of child shapes when increasing "concavity" and "connectionDist" does not help (but some detail gets lost). Slow? Maybe, but can speed up the decomposition process.
	float decimationDistanceInAabbHalfExtentsUnits;				// (0.1). In (0,1]. Bigger results in bigger decimation (= possibly less child shapes)
	bool decimationDistanceUniformInXYZ;						// (true). When true: DDx = DDy = DDz = decimationDistanceInHalfAabbExtentsUnits * min(AabbHalfExtents)xyz;
																//		   When false:  DDx = decimationDistanceInHalfAabbExtentsUnits * AabbHalfExtents.x;	
																//						DDy = decimationDistanceInHalfAabbExtentsUnits * AabbHalfExtents.y;	
																//						DDz = decimationDistanceInHalfAabbExtentsUnits * AabbHalfExtents.z;	
	bool decomposeACenteredCopyOfTheMesh;						// (false). The mesh center is always calculated keeping into account all the subparts of the whole btStridingMeshInterface.
	btVector3 decomposeATranslatedCopyOfTheMesh;				// btVector3(0,0,0). In unscaled units. The center of mass will be shifted in the opposite way.
	btVector3 decomposeAScaledCopyOfTheMesh;					// btVector3(1,1,1).
        std::string optionalBCSSaveFilePath;						// (""). when set, the Bullet Collision Shape is saved as btCompoundShape to disk (*).
	float convexHullsCollisionMargin;							// (0.01f) 
	bool reduceHullVerticesUsingBtShapeHull;					// (false) when set, uses btShapeHull (implemented by John McCutchan) to simplify the original btConvexHullShapes (from HACD decomposition), so that the number of hull vertices in each child shape is less than 42. It should result in a lower number of vertices per hull (when this does not happen, the original btConvexHullShape is used).
	bool convexHullsEnablePolyhedralContactClipping;			// (false) better quality, slightly slower
	bool shrinkObjectInwardsToCompensateCollisionMargin;		// (false)	// Based on the code in appConvexDecompositionDemo. Slow. And seems to produce artifacts...
	bool useStanMelaxVolumeIntegrationForClusterOrigins;		// (false).	If true the child shape transform origins are not placed in the child shape aabb centers, but in their (uniform volume based)
																//			center of mass. However please note that usually child shapes overlap: so this overhead does not make sense in most cases.
	bool storeAdditionalInfoOfChildShapes;						// (false). When set, after the decomposition, an array can be retrieved with info about vertices, indices, volumes (using Stan Melax formula)
																//			of the child shapes (it can't be serialized).
	bool dontCreateAnyChildShape;								// (false). Sometimes useful when 'storeAdditionalInfoOfChildShapes' is true, and we want to create child shapes manually.																 																

	// (*) To load a btCollisionShape saved to disk:
	/*
	#include <BulletWorldImporter/btBulletWorldImporter.h>		// In "Bullet/Extra/Serialize". Needs link to: BulletWorldImporter.lib
	btCollisionShape* Load(const char* filename,bool verbose)	{	
		btBulletWorldImporter loader(0);//don't store info into the world
		loader.setVerboseMode(verbose);
		if (!loader.loadFile(filename)) return NULL;
	
		btCollisionShape* shape = NULL;
		if (loader.getNumCollisionShapes()>0) shape = loader.getCollisionShapeByIndex(0);
	
		//TODO: Cleaner way:
		// 1) Deep clone Collision Shape	
		// 2) loader.deleteAllData(); // deletes all (collision shapes included)
		// 3) return Deep cloned Collision Shape
	
		// Here we don't delete all data from the loader. (leaks?)
		return shape;	
	}	
	*/
	void reset() {*this = Params();}
	void resetForSpeed()	{
		reset();
        depth=5;
        maxConcavity=0.05;
        //invertInputFaces=false;
        posSampling=6;
        angleSampling=6;
        posRefine=3;
        angleRefine=3;
        alpha=0.01;
		targetNTrianglesDecimatedMesh = 500;
        decomposeACleanCopyOfTheMesh = false;
	}
    void resetWithVHACDDefaults()	{
		reset();
        depth=10;
        maxConcavity=0.01;
        //invertInputFaces=false;
        posSampling=10;
        angleSampling=10;
        posRefine=5;
        angleRefine=5;
        alpha=0.01;
		targetNTrianglesDecimatedMesh = 1000;
		decomposeACleanCopyOfTheMesh = false;
	}
    // This setting should produce (sub-optimal but) good results in most cases:
    void resetForCommonUsage()	{
        reset();
        depth=8;
        maxConcavity=0.02;
        //invertInputFaces=false;
        posSampling=10;
        angleSampling=10;
        posRefine=5;
        angleRefine=5;
        alpha=0.01;
        targetNTrianglesDecimatedMesh = 1000;
        decomposeACleanCopyOfTheMesh = true;
    }

	void display(bool modifiedParamsOnly=true)	{
		Params d;
        if (!modifiedParamsOnly || d.depth!=depth) 	printf("params.depth	=				%1d\n",(int) depth);
        if (!modifiedParamsOnly || d.maxConcavity!=maxConcavity) 				printf("params.maxConcavity	=				%1.6f\n",maxConcavity);
        /*if (!modifiedParamsOnly || d.invertInputFaces!=invertInputFaces) {
            if (invertInputFaces) 									printf("params.invertInputFaces	=			true\n");
            else                                                    printf("params.invertInputFaces	=			false\n");
        }*/
        if (!modifiedParamsOnly || d.posSampling!=posSampling)      printf("params.posSampling	=				%1d\n",(int) posSampling);
        if (!modifiedParamsOnly || d.angleSampling!=angleSampling)      printf("params.angleSampling	=				%1d\n",(int) angleSampling);
        if (!modifiedParamsOnly || d.posRefine!=posRefine)      printf("params.posRefine	=				%1d\n",(int) posRefine);
        if (!modifiedParamsOnly || d.angleRefine!=angleRefine)      printf("params.angleRefine	=				%1d\n",(int) angleRefine);
        if (!modifiedParamsOnly || d.alpha!=alpha) 	printf("params.alpha	=				%1d\n",(int) alpha);

        if (!modifiedParamsOnly || d.targetNTrianglesDecimatedMesh!=targetNTrianglesDecimatedMesh) 	printf("params.targetNTrianglesDecimatedMesh=			%d\n",(int)targetNTrianglesDecimatedMesh);
        //if (!modifiedParamsOnly || d.optionalVRMLSaveFilePath!=optionalVRMLSaveFilePath) 	printf("params.optionalVRMLSaveFilePath=			%s\n",optionalVRMLSaveFilePath.c_str());
        if (!modifiedParamsOnly || d.displayDebugInfo!=displayDebugInfo) {
			if (displayDebugInfo) 										printf("params.displayDebugInfo	=				true\n");
			else 														printf("params.displayDebugInfo	=				false\n");
		}		

		if (!modifiedParamsOnly || d.decomposeOnlySelectedSubmeshes.size()!=decomposeOnlySelectedSubmeshes.size()) 	{
			printf ("params.decomposeOnlySelectedSubmeshes	=		");
			if (decomposeOnlySelectedSubmeshes.size()==0) printf("No, thanks\n");
			else	{
				printf("(");
				for (int i=0,sz=decomposeOnlySelectedSubmeshes.size();i<sz;i++)	{
					printf("%1d",decomposeOnlySelectedSubmeshes[i]);
					if (i!=sz-1) printf(",");
				}
				printf(")\n");	
			}	
		}
		if (!modifiedParamsOnly || d.keepSubmeshesSeparated!=keepSubmeshesSeparated) {
			if (keepSubmeshesSeparated) 									printf("params.keepSubmeshesSeparated	=			true\n");
			else 														printf("params.keepSubmeshesSeparated	=			false\n");
		}		
        /*
        if (!modifiedParamsOnly || d.keepSubmeshesSeparatedNClusters.size()!=keepSubmeshesSeparatedNClusters.size()) 	{
			printf ("params.keepSubmeshesSeparatedNClusters	=		");
			if (keepSubmeshesSeparatedNClusters.size()==0) printf("No, thanks\n");
			else	{
				printf("(");
				for (int i=0,sz=keepSubmeshesSeparatedNClusters.size();i<sz;i++)	{
					printf("%1d",keepSubmeshesSeparatedNClusters[i]);
					if (i!=sz-1) printf(",");
				}
				printf(")\n");	
			}	
		}
		if (!modifiedParamsOnly || d.keepSubmeshesSeparatedMaxHullVertices.size()!=keepSubmeshesSeparatedMaxHullVertices.size()) 	{
			printf ("params.keepSubmeshesSeparatedMaxHullVertices	=		");
			if (keepSubmeshesSeparatedMaxHullVertices.size()==0) printf("No, thanks\n");
			else	{
				printf("(");
				for (int i=0,sz=keepSubmeshesSeparatedMaxHullVertices.size();i<sz;i++)	{
					printf("%1d",keepSubmeshesSeparatedMaxHullVertices[i]);
					if (i!=sz-1) printf(",");
				}
				printf(")\n");	
			}	
		}
        */
        if (!modifiedParamsOnly || d.decomposeACleanCopyOfTheMesh!=decomposeACleanCopyOfTheMesh) {
			if (decomposeACleanCopyOfTheMesh) 							printf("params.decomposeACleanCopyOfTheMesh	=		true\n");
			else 														printf("params.decomposeACleanCopyOfTheMesh	=		false\n");
		}		
		if (!modifiedParamsOnly || d.decomposeADecimatedCopyOfTheMesh!=decomposeADecimatedCopyOfTheMesh) {
			if (decomposeADecimatedCopyOfTheMesh) 						printf("params.decomposeADecimatedCopyOfTheMesh	=		true\n");
			else 														printf("params.decomposeADecimatedCopyOfTheMesh	=		false\n");
		}		
		if (!modifiedParamsOnly || d.decimationDistanceInAabbHalfExtentsUnits!=decimationDistanceInAabbHalfExtentsUnits) 	
																		printf("params.decimationDistanceInAabbHalfExtentsUnits	=	%1.6f\n",decimationDistanceInAabbHalfExtentsUnits);		
		if (!modifiedParamsOnly || d.decimationDistanceUniformInXYZ!=decimationDistanceUniformInXYZ) {
			if (decimationDistanceUniformInXYZ) 						printf("params.decimationDistanceUniformInXYZ	=		true\n");
			else 														printf("params.decimationDistanceUniformInXYZ	=		false\n");
		}		
		if (!modifiedParamsOnly || d.decomposeACenteredCopyOfTheMesh!=decomposeACenteredCopyOfTheMesh) {
			if (decomposeACenteredCopyOfTheMesh) 						printf("params.decomposeACenteredCopyOfTheMesh	=		true\n");
			else 														printf("params.decomposeACenteredCopyOfTheMesh	=		false\n");
		}
		if (!modifiedParamsOnly || d.decomposeATranslatedCopyOfTheMesh!=decomposeATranslatedCopyOfTheMesh) 		
																		printf("params.decomposeATranslatedCopyOfTheMesh	=		(%1.3f,%1.3f,%1.3f)\n",decomposeATranslatedCopyOfTheMesh.x(),decomposeATranslatedCopyOfTheMesh.y(),decomposeATranslatedCopyOfTheMesh.z());
		if (!modifiedParamsOnly || d.decomposeAScaledCopyOfTheMesh!=decomposeAScaledCopyOfTheMesh) 		
																		printf("params.decomposeAScaledCopyOfTheMesh	=		(%1.3f,%1.3f,%1.3f)\n",decomposeAScaledCopyOfTheMesh.x(),decomposeAScaledCopyOfTheMesh.y(),decomposeAScaledCopyOfTheMesh.z());
		if (!modifiedParamsOnly || optionalBCSSaveFilePath.size()>0) 				printf("params.optionalBCSSaveFilePath	=			\"%s\"\n",optionalBCSSaveFilePath.c_str());
		if (!modifiedParamsOnly || d.convexHullsCollisionMargin!=convexHullsCollisionMargin) 		printf("params.convexHullsCollisionMargin	=		%1.3f\n",convexHullsCollisionMargin);
		if (!modifiedParamsOnly || d.reduceHullVerticesUsingBtShapeHull!=reduceHullVerticesUsingBtShapeHull) {
			if (reduceHullVerticesUsingBtShapeHull) 			printf("params.reduceHullVerticesUsingBtShapeHull	=			true\n");
			else 												printf("params.reduceHullVerticesUsingBtShapeHull	=			false\n");
		}	
		if (!modifiedParamsOnly || d.convexHullsEnablePolyhedralContactClipping!=convexHullsEnablePolyhedralContactClipping) {
			if (convexHullsEnablePolyhedralContactClipping) 			printf("params.convexHullsEnablePolyhedralContactClipping	=			true\n");
			else 														printf("params.convexHullsEnablePolyhedralContactClipping	=			false\n");
		}	
		if (!modifiedParamsOnly || d.shrinkObjectInwardsToCompensateCollisionMargin!=shrinkObjectInwardsToCompensateCollisionMargin) {
			if (shrinkObjectInwardsToCompensateCollisionMargin) 		printf("params.shrinkObjectInwardsToCompensateCollisionMargin=	true\n");
			else 														printf("params.shrinkObjectInwardsToCompensateCollisionMargin=	false\n");
		}												
		if (!modifiedParamsOnly || d.useStanMelaxVolumeIntegrationForClusterOrigins!=useStanMelaxVolumeIntegrationForClusterOrigins) {
			if (useStanMelaxVolumeIntegrationForClusterOrigins) 		printf("params.useStanMelaxVolumeIntegrationForClusterOrigins=	true\n");
			else 														printf("params.useStanMelaxVolumeIntegrationForClusterOrigins=	false\n");
		}												
		if (!modifiedParamsOnly || d.storeAdditionalInfoOfChildShapes!=storeAdditionalInfoOfChildShapes) {
			if (storeAdditionalInfoOfChildShapes) 						printf("params.storeAdditionalInfoOfChildShapes=	true\n");
			else 														printf("params.storeAdditionalInfoOfChildShapes=	false\n");
		}												
		if (!modifiedParamsOnly || d.dontCreateAnyChildShape!=dontCreateAnyChildShape) {
			if (dontCreateAnyChildShape) 								printf("params.dontCreateAnyChildShape=	true\n");
			else 														printf("params.dontCreateAnyChildShape=	false\n");
		}												
	}
	
	//============
	// DEFAULTS:
	//============
    Params() : 	depth(                              10),
                maxConcavity(                       0.01),
                //invertInputFaces(					false),
                posSampling(						10),
                angleSampling(                      10),
                posRefine(                          5),
                angleRefine(                        5),
                alpha(                              0.01),
                targetNTrianglesDecimatedMesh(		1000),
                //optionalVRMLSaveFilePath(           ""),
                displayDebugInfo(					false),
                keepSubmeshesSeparated(				false),
				decomposeACleanCopyOfTheMesh(		true),
				decomposeADecimatedCopyOfTheMesh(	false),
				decimationDistanceInAabbHalfExtentsUnits(0.1f),
				decimationDistanceUniformInXYZ(		true),
				decomposeACenteredCopyOfTheMesh(	false),
				decomposeATranslatedCopyOfTheMesh(	btVector3(0,0,0)),
				decomposeAScaledCopyOfTheMesh(		btVector3(1,1,1)),
				optionalBCSSaveFilePath(			""),				
				convexHullsCollisionMargin(			0.01f),
				reduceHullVerticesUsingBtShapeHull(	false),
				convexHullsEnablePolyhedralContactClipping(false),
				shrinkObjectInwardsToCompensateCollisionMargin(false),
				useStanMelaxVolumeIntegrationForClusterOrigins(false),
				storeAdditionalInfoOfChildShapes(false),
				dontCreateAnyChildShape(false)
				{}
};

// Main ctr
btVHACDCompoundShape(const btStridingMeshInterface* stridingMeshInterface,const Params& params=Params())	{
	init(stridingMeshInterface,params);
}
/* Creates a decomposed shape of the input collision shape that must be one of the following:
	btBvhTriangleMeshShape
	btScaledBvhTriangleMeshShape
	btGImpactShape
   btCompoundShapes with one nested compatible shape are not supported.
*/
btVHACDCompoundShape(const btCollisionShape* shape,const btVHACDCompoundShape::Params& VHACDparams=btVHACDCompoundShape::Params(),const bool considerShapeLocalScaling=true)	{
			if (!shape) return;
			const btStridingMeshInterface* smi(NULL);
            btVHACDCompoundShape::Params params(VHACDparams);
			if (shape->getShapeType()==GIMPACT_SHAPE_PROXYTYPE)	{
				const btGImpactShapeInterface* gimi = static_cast < const btGImpactShapeInterface* > (shape);
				if (gimi && gimi->getGImpactShapeType()==CONST_GIMPACT_TRIMESH_SHAPE)	{
					const btGImpactMeshShape* gishape = static_cast < const btGImpactMeshShape* > (gimi);
					smi = gishape->getMeshInterface();
					if (considerShapeLocalScaling) params.decomposeAScaledCopyOfTheMesh*=gishape->getLocalScaling();
				}
			}
			else if (shape->getShapeType()==TRIANGLE_MESH_SHAPE_PROXYTYPE)	{
				const btBvhTriangleMeshShape* bvhShape = static_cast < const btBvhTriangleMeshShape* > (shape);
					smi = bvhShape->getMeshInterface();
					if (considerShapeLocalScaling) params.decomposeAScaledCopyOfTheMesh*=bvhShape->getLocalScaling();
			}
			else if (shape->getShapeType()==SCALED_TRIANGLE_MESH_SHAPE_PROXYTYPE)	{
				const btScaledBvhTriangleMeshShape* sbvhShape = static_cast < const btScaledBvhTriangleMeshShape* > (shape);
				if (sbvhShape)	{
					const btBvhTriangleMeshShape* bvhShape = static_cast < const btBvhTriangleMeshShape* > (sbvhShape->getChildShape());
					if (bvhShape)	{
						smi = bvhShape->getMeshInterface();
						params.decomposeAScaledCopyOfTheMesh*=sbvhShape->getLocalScaling();		//This must be left outside 'considerShapeLocalScaling'
						if (considerShapeLocalScaling) params.decomposeAScaledCopyOfTheMesh*=bvhShape->getLocalScaling();
					}	
				}	
			}
			if (!smi) return;
			init(smi,params);
}
virtual ~btVHACDCompoundShape() {}

int getSubmeshIndexOfChildShape(int childShape) const {
	return (childShape>=0 && m_submeshIndexOfChildShapes.size()>childShape) ? m_submeshIndexOfChildShapes[childShape] : -1;
}
const btAlignedObjectArray< int >& getSubmeshIndexOfChildShapesMap() const {return m_submeshIndexOfChildShapes;}
btAlignedObjectArray< int >& getSubmeshIndexOfChildShapesMap() {return m_submeshIndexOfChildShapes;}
void getSubmeshIndexOfChildShapesMapCopy(btAlignedObjectArray< int >& out) const {
	out.copyFromArray(m_submeshIndexOfChildShapes);
}

inline static btVHACDCompoundShape* upcast(btCollisionShape* c)	{
    return dynamic_cast < btVHACDCompoundShape* > (c);
}
inline static const btVHACDCompoundShape* upcast(const btCollisionShape* c)	{
    return dynamic_cast < const btVHACDCompoundShape* > (c);
}

struct AdditionalInfoOfChildShape	{    
	btAlignedObjectArray< btVector3 > verts;	// around 'origin' (they depend on it)
	btAlignedObjectArray< int > 	  inds;

    //In VHACD verts and inds of the concave shard are available
    //If somebody needs them here they are:
    btAlignedObjectArray< btVector3 > vertsConcave;	// around 'origin' (they depend on it)
    btAlignedObjectArray< int > 	  indsConcave;

	btVector3 origin;	// the child transform origin (the child transform basis is always the identity matrix): affected by 'params.useStanMelaxVolumeIntegrationForClusterOrigins'.
	bool originIsUniformVolumeCenter; 	// true if 'params.useStanMelaxVolumeIntegrationForClusterOrigins==true'
	btScalar volume;
	btScalar totalVolumeFraction;
	int submeshIndex;	// >=0 only if:		 params.keepSubmeshesSeparated == true
	
	btVector3 aabbHalfExtents;
	btVector3 aabbCenter;			// When 'params.useStanMelaxVolumeIntegrationForClusterOrigins==false' -> origin = aabbCenter
	btVector3 uniformVolumeCenter;	// When 'params.useStanMelaxVolumeIntegrationForClusterOrigins==true' -> origin = uniformVolumeCenter
	
	AdditionalInfoOfChildShape() : submeshIndex(-1),volume(0),origin(0,0,0),totalVolumeFraction(0),originIsUniformVolumeCenter(false),
								   aabbHalfExtents(0,0,0),aabbCenter(0,0,0),uniformVolumeCenter(0,0,0) {}
	

	static btScalar SumVolumesInArray(const btAlignedObjectArray < AdditionalInfoOfChildShape >& array)	{
		btScalar sum(0.);
		for (int i=0,sz=array.size();i<sz;i++) sum+=array[i].volume;
		return sum;
	}
	static btScalar RecalculateTotalVolumeFractionsInArray(btAlignedObjectArray < AdditionalInfoOfChildShape >& array)	{
		const btScalar totVolume = SumVolumesInArray(array);
		for (int i=0,sz=array.size();i<sz;i++) {
			array[i].totalVolumeFraction = totVolume==0 ? 0 : (array[i].volume/totVolume);
		}	
		return totVolume;
	}
	//! Doesn't update totalVolumeFraction:
	void recalculateVolume(bool recalculateUniformVolumeBasedOrigin = false)	{
        btVHACDCompoundShape::PerformStanMelaxVolumeIntegration(verts,inds,volume,&uniformVolumeCenter);
		if (recalculateUniformVolumeBasedOrigin) origin = uniformVolumeCenter;
	}
	void displayDebugInfo() const {
		printf("verts.size()=%d inds.size()=%d\n",verts.size(),inds.size());
		{const btVector3& v(origin);printf("origin=(%1.3f,%1.3f,%1.3f)\n",v.x(),v.y(),v.z());}
		{const bool b(originIsUniformVolumeCenter);if (b) printf("originIsUniformVolumeCenter = true\n");else printf("originIsUniformVolumeCenter = false\n");}
		printf("volume=%1.3f totalVolumeFraction=%1.3f submeshIndex=%d\n",volume,totalVolumeFraction,submeshIndex);
		{const btVector3& v(aabbHalfExtents);printf("aabbHalfExtents=(%1.3f,%1.3f,%1.3f)\n",v.x(),v.y(),v.z());}
		{const btVector3& v(aabbCenter);printf("aabbCenter=(%1.3f,%1.3f,%1.3f)\n",v.x(),v.y(),v.z());}
		{const btVector3& v(uniformVolumeCenter);printf("uniformVolumeCenter=(%1.3f,%1.3f,%1.3f)\n",v.x(),v.y(),v.z());}		
	}	
};
typedef btAlignedObjectArray < AdditionalInfoOfChildShape > AdditionalInfoOfChildShapes;

// Available only if 'params.storeAdditionalInfoOfChildShapes == true':
const AdditionalInfoOfChildShape* getAdditionalInfoOfChildShape(int childShape) const {return (childShape>=0 && m_additionalInfoOfChildShapes.size()>childShape) ? &(m_additionalInfoOfChildShapes[childShape]) : NULL;}
const AdditionalInfoOfChildShapes& getAdditionalInfoOfChildShapes() const {return m_additionalInfoOfChildShapes;}
AdditionalInfoOfChildShapes& getAdditionalInfoOfChildShapes() {return m_additionalInfoOfChildShapes;}
void getAdditionalInfoOfChildShapesCopy(AdditionalInfoOfChildShapes& out) const {
	out.copyFromArray(m_additionalInfoOfChildShapes);
}

protected:
//for easy inheriting:
btVHACDCompoundShape() {}

// Main Method:
void init(const btStridingMeshInterface* stridingMeshInterface,const Params& params=Params()) {
	m_submeshIndexOfChildShapes.resize(0);
	if (!stridingMeshInterface) return;
	
	const std::clock_t ck1 = std::clock();
        btVHACDCompoundShape::Params::IntVector decomposeOnlySelectedSubmeshes = params.decomposeOnlySelectedSubmeshes;
	if (decomposeOnlySelectedSubmeshes.size()) {
                #ifndef BTHACDCOMPOUNDSHAPE_H_USES_STDVECTOR_AS_VECTOR
                decomposeOnlySelectedSubmeshes.quickSort(Less());
                #else   //BTHACDCOMPOUNDSHAPE_H_USES_STDVECTOR_AS_VECTOR
		std::sort(decomposeOnlySelectedSubmeshes.begin(),decomposeOnlySelectedSubmeshes.end(),Less());
                //#undef BTHACDCOMPOUNDSHAPE_H_USES_STDVECTOR_AS_VECTOR      // No undef, even if it's no more needed...
                #endif  //BTHACDCOMPOUNDSHAPE_H_USES_STDVECTOR_AS_VECTOR
	}	

    const VHACD::Vec3<VHACD::Real> meshTranslation = VHACD::Vec3<VHACD::Real>(params.decomposeATranslatedCopyOfTheMesh.x(),params.decomposeATranslatedCopyOfTheMesh.y(),params.decomposeATranslatedCopyOfTheMesh.z());
	const bool mustTranslateMesh = (params.decomposeATranslatedCopyOfTheMesh.x()!=0 || params.decomposeATranslatedCopyOfTheMesh.y()!=0 || params.decomposeATranslatedCopyOfTheMesh.z()!=0);
    const VHACD::Vec3<VHACD::Real> meshScalingFactor = VHACD::Vec3<VHACD::Real>(params.decomposeAScaledCopyOfTheMesh.x(),params.decomposeAScaledCopyOfTheMesh.y(),params.decomposeAScaledCopyOfTheMesh.z());
	const bool mustScaleMesh = (params.decomposeAScaledCopyOfTheMesh.x()!=1 || params.decomposeAScaledCopyOfTheMesh.y()!=1 || params.decomposeAScaledCopyOfTheMesh.z()!=1);
    VHACD::Vec3<VHACD::Real> aabbHalfExtents;
    VHACD::Vec3<VHACD::Real> aabbCenterPoint;
	if (params.decomposeADecimatedCopyOfTheMesh || params.decomposeACenteredCopyOfTheMesh)	{
		aabbHalfExtents = GetAabbHalfExtentsFromStridingInterface(stridingMeshInterface,&aabbCenterPoint);//,meshScalingFactor);
		aabbHalfExtents = aabbHalfExtents*meshScalingFactor;
	}

    VHACD::Vec3<VHACD::Real> decimationDistance(SIMD_EPSILON,SIMD_EPSILON,SIMD_EPSILON);
	if (params.decomposeADecimatedCopyOfTheMesh)	{
		if (params.decimationDistanceUniformInXYZ) {
            const VHACD::Real DD = params.decimationDistanceInAabbHalfExtentsUnits * (
																					(aabbHalfExtents.X() < aabbHalfExtents.Y() && aabbHalfExtents.X() < aabbHalfExtents.Z()) ? aabbHalfExtents.X() :
																					(aabbHalfExtents.Y() < aabbHalfExtents.X() && aabbHalfExtents.Y() < aabbHalfExtents.Z()) ? aabbHalfExtents.Y() :
																					 aabbHalfExtents.Z()
																					 );
            decimationDistance = VHACD::Vec3<VHACD::Real>(DD,DD,DD);
		}																					 
        else decimationDistance = VHACD::Vec3<VHACD::Real>(params.decimationDistanceInAabbHalfExtentsUnits*aabbHalfExtents.X(),params.decimationDistanceInAabbHalfExtentsUnits*aabbHalfExtents.Y(),params.decimationDistanceInAabbHalfExtentsUnits*aabbHalfExtents.Z());
	}	
	

	if (params.displayDebugInfo) printf("NumSubmeshes: %1d\n",stridingMeshInterface->getNumSubParts());
	const bool forceDegenerateTrianglesRemovalWhenACleanCopyOfTheMeshMustBeProcessed = true;	// Yes, much better
	
	
    std::vector< VHACD::Vec3<VHACD::Real> > points;
    std::vector< VHACD::Vec3<long> > triangles;
		
	if (!params.keepSubmeshesSeparated)	{
		GetStridingInterfaceContent(stridingMeshInterface,points,triangles,decomposeOnlySelectedSubmeshes.size()>0 ? &decomposeOnlySelectedSubmeshes : NULL);
		if (params.displayDebugInfo) printf("Numverts: %1d NumTriangles: %1d (from btStridingMeshInterface)\n",(int) points.size(),(int)triangles.size());	
		if (params.decomposeACenteredCopyOfTheMesh) ShiftPoints(points,-aabbCenterPoint);
		if (mustTranslateMesh) ShiftPoints(points,meshTranslation);
		if (mustScaleMesh) ScalePoints(points,meshScalingFactor);

		if (params.decomposeADecimatedCopyOfTheMesh || params.decomposeACleanCopyOfTheMesh)	{
			const bool removeDegenerateTrianglesToo = params.decomposeADecimatedCopyOfTheMesh | forceDegenerateTrianglesRemovalWhenACleanCopyOfTheMeshMustBeProcessed; // Usually when removing double vertices, degenerate triangles are less prone to appear...
			RemoveDoubleVertices(points,triangles,removeDegenerateTrianglesToo,decimationDistance);	// This same method is used to decimate the mesh too.
			if (params.displayDebugInfo)	{
				if (params.decomposeADecimatedCopyOfTheMesh) printf("Numverts: %1d NumTriangles: %1d (after mesh decimation)\n",(int) points.size(),(int)triangles.size());	
				else if (params.decomposeACleanCopyOfTheMesh)  printf("Numverts: %1d NumTriangles: %1d (after duplicated vertices removal)\n",(int) points.size(),(int)triangles.size());	
			}	
		}
        if (params.targetNTrianglesDecimatedMesh>0 && params.targetNTrianglesDecimatedMesh < triangles.size()) {
            VHACD::MeshDecimator myMDecimator;
            myMDecimator.SetCallBack(&btVHACDCompoundShape::VHACDCallBackFunction);
            myMDecimator.Initialize(points.size(), triangles.size(), &points[0], &triangles[0]);
            myMDecimator.Decimate(0, params.targetNTrianglesDecimatedMesh);

            points.resize(myMDecimator.GetNVertices());
            triangles.resize(myMDecimator.GetNTriangles());
            if (points.size()>0 && triangles.size()>0) myMDecimator.GetMeshData(&points[0], &triangles[0]);
            else {points.clear();triangles.clear();}
        }

		performHACDMainWork(params,points,triangles);
	}
	else	{
		const int subParts = stridingMeshInterface->getNumSubParts();
		int curSelSm = -1;
		int curSelSmIdx = 0;	
		const bool mustSelectSubMeshes = (decomposeOnlySelectedSubmeshes.size()>0);			
		if (mustSelectSubMeshes) curSelSm = decomposeOnlySelectedSubmeshes[curSelSmIdx];
        std::vector< VHACD::Vec3<VHACD::Real> > points;
        std::vector< VHACD::Vec3<long> > triangles;
		for (int subPart = 0;subPart<subParts;subPart++)	{
			if (mustSelectSubMeshes)	{
				if (curSelSm!=subPart) continue;
				if (decomposeOnlySelectedSubmeshes.size()<=curSelSmIdx+1) curSelSm = subParts;	// This will never process successive cycles
				else curSelSm =  decomposeOnlySelectedSubmeshes[++curSelSmIdx];					// Ready for next cycle	
			}
			points.clear();
			triangles.clear();
			GetStridingInterfaceContentOfSingleSubmesh(stridingMeshInterface,points,triangles,subPart);
		
			if (params.displayDebugInfo) {
				printf("Subpart: %1d. Numverts: %1d NumTriangles: %1d (From btStridingMeshInterface)\n",subPart,(int) points.size(),(int)triangles.size());	
			}
			
			if (params.decomposeACenteredCopyOfTheMesh) ShiftPoints(points,-aabbCenterPoint);
			if (mustTranslateMesh) ShiftPoints(points,meshTranslation);
			if (mustScaleMesh) ScalePoints(points,meshScalingFactor);
			
			
			if (params.decomposeADecimatedCopyOfTheMesh || params.decomposeACleanCopyOfTheMesh)	{
				const bool removeDegenerateTrianglesToo = params.decomposeADecimatedCopyOfTheMesh | forceDegenerateTrianglesRemovalWhenACleanCopyOfTheMeshMustBeProcessed; // Usually when removing double vertices, degenerate triangles are less prone to appear...
				RemoveDoubleVertices(points,triangles,removeDegenerateTrianglesToo,decimationDistance);	// This same method is used to decimate the mesh too.
				if (params.displayDebugInfo)	{
					if (params.decomposeADecimatedCopyOfTheMesh) printf("Subpart: %1d. Numverts: %1d NumTriangles: %1d (after mesh decimation)\n",subPart,(int) points.size(),(int)triangles.size());	
					else if (params.decomposeACleanCopyOfTheMesh)  printf("Subpart: %1d. Numverts: %1d NumTriangles: %1d (after duplicated vertices removal)\n",subPart,(int) points.size(),(int)triangles.size());	
				}
			}
	
            if (params.targetNTrianglesDecimatedMesh>0 && params.targetNTrianglesDecimatedMesh < triangles.size()) {
                VHACD::MeshDecimator myMDecimator;
                myMDecimator.SetCallBack(&btVHACDCompoundShape::VHACDCallBackFunction);
                myMDecimator.Initialize(points.size(), triangles.size(), &points[0], &triangles[0]);
                myMDecimator.Decimate(0, params.targetNTrianglesDecimatedMesh);

                points.resize(myMDecimator.GetNVertices());
                triangles.resize(myMDecimator.GetNTriangles());
                if (points.size()>0 && triangles.size()>0) myMDecimator.GetMeshData(&points[0], &triangles[0]);
                else {points.clear();triangles.clear();}
            }

			performHACDMainWork(params,points,triangles,subPart);
		}			
		for (int i=0,sz=m_submeshIndexOfChildShapes.size();i<sz;i++)	{
			printf("submeshIndexOfChildShapes[%1d]	=	%1d;\n",i,m_submeshIndexOfChildShapes[i]);		
		}
	}	
	
	if (params.optionalBCSSaveFilePath.size()>0 && this->getNumChildShapes()>0)	{
		 	btDefaultSerializer serializer;
 			serializer.startSerialization();
			this->serializeSingleShape(&serializer);
			serializer.finishSerialization();  	
			FILE* file = fopen(params.optionalBCSSaveFilePath.c_str(),"wb");
			if (file)	{
				fwrite(serializer.getBufferPointer(),serializer.getCurrentBufferSize(),1, file);
   				fclose(file);
   				if (params.displayDebugInfo) printf("btCompoundShape saved to: \"%s\"\n",params.optionalBCSSaveFilePath.c_str());
   			}
   			else {	
   				 printf("ERROR: I can't save the btCompoundShape to: \"%s\"\n",params.optionalBCSSaveFilePath.c_str());
   			}
			
	}
	const std::clock_t ck2 = std::clock();
    if (params.displayDebugInfo) printf("Operation performed in %d seconds.\n",(int)( ((double)(ck2-ck1))/(double)CLOCKS_PER_SEC) );
}



void performHACDMainWork(const Params& params,std::vector< VHACD::Vec3<VHACD::Real> >& points,std::vector< VHACD::Vec3<long> >& triangles,const int subPart=-1)	{

    VHACD::Mesh mesh;
    size_t nPoints = points.size(),nTriangles=triangles.size();
    mesh.ResizePoints(nPoints);
    mesh.ResizeTriangles(nTriangles);
    for(size_t p = 0; p < nPoints;    ++p) mesh.SetPoint(p, points[p]);
    for(size_t t = 0; t < nTriangles; ++t) mesh.SetTriangle(t, triangles[t]);

    std::vector< VHACD::Mesh * > parts;
    VHACD::ApproximateConvexDecomposition(mesh, params.depth, params.posSampling, params.angleSampling, params.posRefine, params.angleRefine, params.alpha, params.maxConcavity, parts, &btVHACDCompoundShape::VHACDCallBackFunction);
    const size_t nParts = parts.size();
    //printf("VHACD::ApproximateConvexDecomposition: numClusters = %d\n",(int)nParts);

        //if (!params.keepSubmeshesSeparated && params.optionalVRMLSaveFilePath.size()>0) mesh.SaveVRML2(params.optionalVRMLSaveFilePath.c_str(), 0);
				
        for (size_t CH = 0,nClusters = nParts; CH < nClusters; ++CH)	{
            VHACD::Mesh* pm = parts[CH];
            if (!pm) continue;
            VHACD::Mesh& concaveShard = *pm;
            VHACD::Mesh m;  // convex hull of concaveShard

            //#define USE_CONCAVE_VERTS_DIRECTLY    // in calculations (and all the concave verts are added to btConvexHullShape). However we lose the convex data that should be passed to additionalInfoStruct if needed.
            #ifdef USE_CONCAVE_VERTS_DIRECTLY
            m=concaveShard;
            #else //USE_CONCAVE_VERTS_DIRECTLY
            concaveShard.ComputeConvexHull(m);
            /*
            // concaveShard.ComputeConvexHull(m); UNWRAPPED:
            // Compute convex hull from concave shard (slightly faster than the 2 line above, expose shrink params):
            {
                //=====================================================================================================
                //======================= Faster (code from: <vhacdMesh.cpp>)==========================================
                btConvexHullComputer ch;
                ch.compute(concaveShard.GetPoints(), 3 * sizeof(VHACD::Real), concaveShard.GetNPoints(), -1.0, -1.0);   // so I get control of the last two params myself, if needed...
                for(int v = 0; v < ch.vertices.size(); v++) {
                    m.AddPoint(VHACD::Vec3<VHACD::Real>(ch.vertices[v].getX(), ch.vertices[v].getY(), ch.vertices[v].getZ()));
                }
                const int nt = ch.faces.size();
                for(int t = 0; t < nt; ++t) {
                    const btConvexHullComputer::Edge * sourceEdge = & (ch.edges[ch.faces[t]]);
                    int a = sourceEdge->getSourceVertex();
                    int b = sourceEdge->getTargetVertex();
                    const btConvexHullComputer::Edge * edge = sourceEdge->getNextEdgeOfFace();
                    int c = edge->getTargetVertex();
                    while (c != a)  {
                        m.AddTriangle(VHACD::Vec3<long>(a, b, c));
                        edge = edge->getNextEdgeOfFace();
                        b = c;
                        c = edge->getTargetVertex();
                    }
                }
            //=====================================================================================================
            }*/
            #endif //USE_CONCAVE_VERTS_DIRECTLY


            const size_t nPoints =  m.GetNPoints();
            const size_t nTriangles =  m.GetNTriangles();
			
            const VHACD::Vec3<VHACD::Real>* points = (const VHACD::Vec3<VHACD::Real>*) m.GetPoints();
            const VHACD::Vec3<long>* triangles = (const VHACD::Vec3<long>*) m.GetTriangles();
            if (points && triangles)	{
				btAlignedObjectArray < btVector3 > verts;
				btVector3 centroid(0,0,0),aabbHalfExtents(0,0,0),uniformVolumeCenter(0,0,0),aabbCenter(0,0,0);
				btScalar volume(0);			// Used only if needed	

				if (!params.dontCreateAnyChildShape || params.storeAdditionalInfoOfChildShapes)	{
					#pragma region Calculate centroid (and store additional info when required)
                    if (!params.useStanMelaxVolumeIntegrationForClusterOrigins || params.storeAdditionalInfoOfChildShapes) aabbHalfExtents = GetAabbHalfExtents(points,nPoints,&aabbCenter);
                    if (params.useStanMelaxVolumeIntegrationForClusterOrigins || params.storeAdditionalInfoOfChildShapes) PerformStanMelaxVolumeIntegration(points,nPoints,triangles,nTriangles,volume,&uniformVolumeCenter);
					
 					if (!params.useStanMelaxVolumeIntegrationForClusterOrigins)	centroid = aabbCenter;			
					else centroid = uniformVolumeCenter;
					#pragma endregion					
					#pragma region Calculate verts array by subtracting centroid
					verts.resize(nPoints);
					for (size_t i=0; i<nPoints; i++)	{
						verts[i]=btVector3(points[i].X()-centroid.x(),points[i].Y()-centroid.y(),points[i].Z()-centroid.z());
					}
					#pragma endregion
				}	
 				if (!params.dontCreateAnyChildShape)	{
					btConvexHullShape* convexHullShape = NULL;					
					#pragma region Calculate convexHullShape							
					if (params.shrinkObjectInwardsToCompensateCollisionMargin)	{
						btAlignedObjectArray<btVector3> planeEquations;
						btGeometryUtil::getPlaneEquationsFromVertices(verts,planeEquations);

						btAlignedObjectArray<btVector3> shiftedPlaneEquations;
						btVector3 plane;
						for (int p=0,psz = planeEquations.size();p<psz;p++)
						{
							plane = planeEquations[p];
							plane[3] += params.convexHullsCollisionMargin;
							shiftedPlaneEquations.push_back(plane);
						}
						btAlignedObjectArray<btVector3> shiftedVertices;
						btGeometryUtil::getVerticesFromPlaneEquations(shiftedPlaneEquations,shiftedVertices);
					
						convexHullShape = new btConvexHullShape(&(shiftedVertices[0].getX()),shiftedVertices.size());				
					}							
					else convexHullShape = new btConvexHullShape(&verts[0].x(),verts.size());
					convexHullShape->setMargin(params.convexHullsCollisionMargin);
					if (params.reduceHullVerticesUsingBtShapeHull)	{
						//create a hull approximation
   						btShapeHull* hull = new btShapeHull(convexHullShape);
						if (hull)	{
							hull->buildHull(params.convexHullsCollisionMargin);
							if (hull->numVertices() < verts.size())	{
								delete convexHullShape;convexHullShape = NULL;
								convexHullShape = new btConvexHullShape((btScalar*)hull->getVertexPointer(),hull->numVertices());
								convexHullShape->setMargin(params.convexHullsCollisionMargin);
							}
   						delete hull;hull = NULL;
   						}	
					}
					if (params.convexHullsEnablePolyhedralContactClipping) convexHullShape->initializePolyhedralFeatures();	
					#pragma endregion			
 					this->addChildShape(btTransform(btQuaternion::getIdentity(),centroid),convexHullShape);	
				} 					
 				
 				if (params.keepSubmeshesSeparated) m_submeshIndexOfChildShapes.push_back(subPart);
 				if (params.storeAdditionalInfoOfChildShapes)	{
 					AdditionalInfoOfChildShape ai;
 					if (params.keepSubmeshesSeparated) ai.submeshIndex = subPart;
 					ai.origin = centroid;
 					ai.verts.copyFromArray(verts);
                    ai.inds.resize(nTriangles*3);
 					int i3;
                    for (int i = 0,trSz = (int)nTriangles;i<trSz;i++)	{
 						i3 = i*3;
                        const VHACD::Vec3<long>& tri(triangles[i]);
 						ai.inds[i3] 	=  tri[0];
 						ai.inds[i3+1] 	=  tri[1];
 						ai.inds[i3+2] 	=  tri[2];
 					}
 					ai.origin = centroid;
 					ai.aabbHalfExtents = aabbHalfExtents;
 					ai.volume = volume;
 					ai.uniformVolumeCenter = uniformVolumeCenter;
 					ai.aabbCenter = aabbCenter;
 					ai.originIsUniformVolumeCenter = params.useStanMelaxVolumeIntegrationForClusterOrigins;

                    //-------- Add Concave Mesh Info -----------------------
                    size_t nPointsConcave = concaveShard.GetNPoints();
                    size_t nTrianglesConcave = concaveShard.GetNTriangles();
                    const VHACD::Vec3<VHACD::Real>* pointsConcave = (const VHACD::Vec3<VHACD::Real>*) concaveShard.GetPoints();
                    const VHACD::Vec3<long>* trianglesConcave = (const VHACD::Vec3<long>*) concaveShard.GetTriangles();
                    ai.vertsConcave.resize(nPointsConcave);
                    for (size_t i=0; i<nPointsConcave; i++)	{
                        ai.vertsConcave[i]=btVector3(pointsConcave[i].X()-centroid.x(),pointsConcave[i].Y()-centroid.y(),pointsConcave[i].Z()-centroid.z());
                        //ai.vertsConcave[i]=btVector3(pointsConcave[i].X(),pointsConcave[i].Y(),pointsConcave[i].Z());
                        //ai.vertsConcave[i]=btVector3(pointsConcave[i].X()+centroid.x(),pointsConcave[i].Y()+centroid.y(),pointsConcave[i].Z()+centroid.z());

                    }
                    ai.indsConcave.resize(nTrianglesConcave*3);
                    for (int i = 0,trSz = (int)nTrianglesConcave;i<trSz;i++)	{
                        i3 = i*3;
                        const VHACD::Vec3<long>& tri(trianglesConcave[i]);
                        ai.indsConcave[i3] 	=  tri[0];
                        ai.indsConcave[i3+1] 	=  tri[1];
                        ai.indsConcave[i3+2] 	=  tri[2];
                    }
                    //------------------------------------------------------

 					m_additionalInfoOfChildShapes.push_back(ai);
 				}					
			}
		}
		
		if (params.storeAdditionalInfoOfChildShapes) AdditionalInfoOfChildShape::RecalculateTotalVolumeFractionsInArray(m_additionalInfoOfChildShapes);


    for(size_t p = 0; p < nParts; ++p)
    {
        //VHACD::Mesh ch;parts[p]->ComputeConvexHull(ch);
        delete parts[p];parts[p]=NULL;
    }
    parts.clear();

}		


virtual const char*	getName()const {return "btVHACDCompoundShape";}

btAlignedObjectArray< int > m_submeshIndexOfChildShapes;
AdditionalInfoOfChildShapes m_additionalInfoOfChildShapes;

static void GetStridingInterfaceContentOfSingleSubmesh(const btStridingMeshInterface* sti,std::vector< VHACD::Vec3<VHACD::Real> >& vertsOut,std::vector< VHACD::Vec3<long> >& trianglesOut,int subMeshIndex)	{
    typedef VHACD::Real vertexType;
	typedef long indexType;

	vertsOut.clear();trianglesOut.clear();
	if (!sti) return;
				
	const int subparts = sti->getNumSubParts();
	if (subparts<=0 || subMeshIndex<0 || subMeshIndex>=subparts) return;		
								
	const unsigned char *pverts;int numVerts;	PHY_ScalarType type;int stride;
	const unsigned char *pinds;int indsStride;int numTriangles;PHY_ScalarType indsType;
	
	unsigned int vCnt=0;
	unsigned int tCnt=0;
		
	const int subpart = subMeshIndex;
						 
	 	{
		sti->getLockedReadOnlyVertexIndexBase(&pverts,numVerts,type,stride,&pinds,indsStride,numTriangles,indsType,subpart);
		const size_t startVerts = vertsOut.size();
		const size_t startTriangles = trianglesOut.size();
		
		try	{
			
							
			switch (type)	{
				case PHY_FLOAT:	{
					typedef float glVertex;
					const glVertex* verts = (const glVertex*) pverts; 
					const int vertexDeltaStrideInGLVertexUnits = stride <= sizeof(glVertex)*3 ? 0 : (stride-sizeof(glVertex)*3)/sizeof(glVertex); 			
					
					vertsOut.resize(startVerts+numVerts);
					for (size_t i = 0; i < numVerts;++i)	{
						vertsOut[vCnt].X()=(vertexType) *verts++;
						vertsOut[vCnt].Y()=(vertexType) *verts++;
						vertsOut[vCnt++].Z()=(vertexType) *verts++;
						
						verts+=vertexDeltaStrideInGLVertexUnits;
					}					
			 	break;
		 		}
				case PHY_DOUBLE:{
					typedef double glVertex;
					const glVertex* verts = (const glVertex*) pverts; 
					const int vertexDeltaStrideInGLVertexUnits = stride <= sizeof(glVertex)*3 ? 0 : (stride-sizeof(glVertex)*3)/sizeof(glVertex); 			

					vertsOut.resize(startVerts+numVerts);
					for (size_t i = 0; i < numVerts;++i)	{
						vertsOut[vCnt].X()=(vertexType) *verts++;
						vertsOut[vCnt].Y()=(vertexType) *verts++;
						vertsOut[vCnt++].Z()=(vertexType) *verts++;
						
						verts+=vertexDeltaStrideInGLVertexUnits;
					}
			 	break;
		 		}
				default:
				btAssert((type == PHY_FLOAT) || (type == PHY_DOUBLE));
			}
		

			switch (indsType)	{
				case PHY_INTEGER:	{
			 		typedef int glIndex;
			 		const glIndex* inds = (const glIndex*) pinds; 
			 		const int indexDeltaStrideInGLIndexUnits = indsStride <= sizeof(indsStride)*3 ? 0 : (stride-sizeof(indsStride)*3)/sizeof(indsStride); 
			 		
					trianglesOut.resize(startTriangles+numTriangles);
					for (size_t i = 0; i < numTriangles;++i)	{
						trianglesOut[tCnt].X() = (indexType) *inds++;
						trianglesOut[tCnt].Y() = (indexType) *inds++;
						trianglesOut[tCnt++].Z() = (indexType) *inds++;
						
						inds+=indexDeltaStrideInGLIndexUnits;
					}
				 	break;
				}
				case PHY_SHORT:	{
					typedef unsigned short glIndex;
			 		const glIndex* inds = (const glIndex*) pinds;
			 		const int indexDeltaStrideInGLIndexUnits = indsStride <= sizeof(indsStride)*3 ? 0 : (stride-sizeof(indsStride)*3)/sizeof(indsStride); 			 		
			 		
					trianglesOut.resize(startTriangles+numTriangles);
					for (size_t i = 0; i < numTriangles;++i)	{
						trianglesOut[tCnt].X() = (indexType) *inds++;
						trianglesOut[tCnt].Y() = (indexType) *inds++;
						trianglesOut[tCnt++].Z() = (indexType) *inds++;
						
						inds+=indexDeltaStrideInGLIndexUnits;												
					}					 		 
				 	break;
				}
				default:
			 	btAssert((indsType == PHY_INTEGER) || (indsType == PHY_SHORT));
			}
		}
		catch (...) {
			vertsOut.resize(startVerts);
			trianglesOut.resize(startTriangles);
		}
		
		sti->unLockReadOnlyVertexBase(subpart);
		}		
}
static void GetStridingInterfaceContent(const btStridingMeshInterface* sti,std::vector< VHACD::Vec3<VHACD::Real> >& vertsOut,std::vector< VHACD::Vec3<long> >& trianglesOut,const btVHACDCompoundShape::Params::IntVector* processOnlySelectedSubmeshesInAscendingOrder=NULL)	{
    typedef VHACD::Real vertexType;
	typedef long indexType;

	vertsOut.clear();trianglesOut.clear();
	if (!sti) return;
				
	const int subparts = sti->getNumSubParts();
	if (subparts<=0 || (processOnlySelectedSubmeshesInAscendingOrder && processOnlySelectedSubmeshesInAscendingOrder->size()==0)) return;		
								
	const unsigned char *pverts;int numVerts;	PHY_ScalarType type;int stride;
	const unsigned char *pinds;int indsStride;int numTriangles;PHY_ScalarType indsType;
	
	btAlignedObjectArray < const unsigned char * > vertexBaseOfProcessedSubMeshes;
	btAlignedObjectArray < int > startVertexIndexOfVertexBaseOfProcessedSubMeshes;
	//#define DONT_REUSE_SHARED_VERTEX_ARRAY	// problably safer, but much slower
	
	unsigned int vCnt=0;
	unsigned int tCnt=0;
		
	int curSelSm = -1;
	int curSelSmIdx = 0;	
	int numProcessedSubmeshes = 0;
	const bool mustSelectSubMeshes = processOnlySelectedSubmeshesInAscendingOrder;			
	if (mustSelectSubMeshes) curSelSm = (*processOnlySelectedSubmeshesInAscendingOrder)[curSelSmIdx];
	
	//::MessageBox(NULL,sw::ToString(subParts).c_str(),_T("subParts"),MB_OK);	
									
	for (int subpart=0;subpart<subparts;subpart++)	{	
		//-----------------------------------------------------------------------------------------------------------------------------------
		if (curSelSm>=0 && curSelSm!=subpart) continue;
		else if (mustSelectSubMeshes) {
			// Prepare the index for next cycle and continue:
			if (processOnlySelectedSubmeshesInAscendingOrder->size()<=curSelSmIdx+1) curSelSm = subparts;	// This will never process successive cycles
			else curSelSm =  (*processOnlySelectedSubmeshesInAscendingOrder)[++curSelSmIdx];				// Ready for next cycle			
		}
		//------------------------------------------------------------------------------------------------------------------------------------				
		sti->getLockedReadOnlyVertexIndexBase(&pverts,numVerts,type,stride,&pinds,indsStride,numTriangles,indsType,subpart);
		const size_t startVerts = vertsOut.size();
		const size_t startTriangles = trianglesOut.size();
		
		int startVertexIndex = startVerts;
		bool vertexArrayIsShared = false;	
		
		#ifndef DONT_REUSE_SHARED_VERTEX_ARRAY	
		{
			for (int i = 0,sz = vertexBaseOfProcessedSubMeshes.size();i<sz;i++)	{
				if (pverts == vertexBaseOfProcessedSubMeshes[i])	{
					startVertexIndex = startVertexIndexOfVertexBaseOfProcessedSubMeshes[i];
					vertexArrayIsShared = true;
					break;
				}
			} 
		
			vertexBaseOfProcessedSubMeshes.push_back(pverts);
			startVertexIndexOfVertexBaseOfProcessedSubMeshes.push_back(startVertexIndex);
		}
		#else //DONT_REUSE_SHARED_VERTEX_ARRAY
			#undef DONT_REUSE_SHARED_VERTEX_ARRAY
		#endif //DONT_REUSE_SHARED_VERTEX_ARRAY
		
		const indexType indsAdder = (indexType) startVertexIndex; 
				
		
		try	{
			
			if (!vertexArrayIsShared)				
			switch (type)	{
				case PHY_FLOAT:	{
					typedef float glVertex;
					const glVertex* verts = (const glVertex*) pverts; 
					const int vertexDeltaStrideInGLVertexUnits = stride <= sizeof(glVertex)*3 ? 0 : (stride-sizeof(glVertex)*3)/sizeof(glVertex); 			
					
					vertsOut.resize(startVerts+numVerts);
					for (size_t i = 0; i < numVerts;++i)	{
						vertsOut[vCnt].X()=(vertexType) *verts++;
						vertsOut[vCnt].Y()=(vertexType) *verts++;
						vertsOut[vCnt++].Z()=(vertexType) *verts++;
						
						verts+=vertexDeltaStrideInGLVertexUnits;
					}					
			 	break;
		 		}
				case PHY_DOUBLE:{
					typedef double glVertex;
					const glVertex* verts = (const glVertex*) pverts; 
					const int vertexDeltaStrideInGLVertexUnits = stride <= sizeof(glVertex)*3 ? 0 : (stride-sizeof(glVertex)*3)/sizeof(glVertex); 			

					vertsOut.resize(startVerts+numVerts);
					for (size_t i = 0; i < numVerts;++i)	{
						vertsOut[vCnt].X()=(vertexType) *verts++;
						vertsOut[vCnt].Y()=(vertexType) *verts++;
						vertsOut[vCnt++].Z()=(vertexType) *verts++;
						
						verts+=vertexDeltaStrideInGLVertexUnits;
					}
			 	break;
		 		}
				default:
				btAssert((type == PHY_FLOAT) || (type == PHY_DOUBLE));
			}
		

			switch (indsType)	{				
				case PHY_INTEGER:	{
			 		typedef int glIndex;
			 		const glIndex* inds = (const glIndex*) pinds; 
			 		const int indexDeltaStrideInGLIndexUnits = indsStride <= sizeof(indsStride)*3 ? 0 : (stride-sizeof(indsStride)*3)/sizeof(indsStride); 
			 		
					trianglesOut.resize(startTriangles+numTriangles);
					for (size_t i = 0; i < numTriangles;++i)	{
						trianglesOut[tCnt].X() = startVertexIndex + (indexType) *inds++;
						trianglesOut[tCnt].Y() = startVertexIndex + (indexType) *inds++;
						trianglesOut[tCnt++].Z() = startVertexIndex + (indexType) *inds++;
						
						inds+=indexDeltaStrideInGLIndexUnits;
					}
				 	break;
				}
				case PHY_SHORT:	{
					typedef unsigned short glIndex;
			 		const glIndex* inds = (const glIndex*) pinds;
			 		const int indexDeltaStrideInGLIndexUnits = indsStride <= sizeof(indsStride)*3 ? 0 : (stride-sizeof(indsStride)*3)/sizeof(indsStride); 			 		
			 		
					trianglesOut.resize(startTriangles+numTriangles);
					for (size_t i = 0; i < numTriangles;++i)	{
						trianglesOut[tCnt].X() = startVertexIndex + (indexType) *inds++;
						trianglesOut[tCnt].Y() = startVertexIndex + (indexType) *inds++;
						trianglesOut[tCnt++].Z() = startVertexIndex + (indexType) *inds++;
						
						inds+=indexDeltaStrideInGLIndexUnits;												
					}					 		 
				 	break;
				}
				default:
			 	btAssert((indsType == PHY_INTEGER) || (indsType == PHY_SHORT));
			}
		}
		catch (...) {
			vertsOut.resize(startVerts);
			trianglesOut.resize(startTriangles);
		}
		
		sti->unLockReadOnlyVertexBase(subpart);
		}	
		
}


static VHACD::Vec3<VHACD::Real> GetAabbHalfExtentsFromStridingInterface(const btStridingMeshInterface* sti,VHACD::Vec3<VHACD::Real>* pOptionalCenterPointOut=NULL,const VHACD::Vec3<VHACD::Real>& scalingFactorToApply=VHACD::Vec3<VHACD::Real>(1,1,1))	{
    typedef VHACD::Real vertexType;
	
    VHACD::Vec3<vertexType> aabbHalfExtents(0,0,0);
    if (pOptionalCenterPointOut) *pOptionalCenterPointOut = VHACD::Vec3<VHACD::Real>(0,0,0);
	
    VHACD::Vec3<vertexType> cmax(0,0,0);
    VHACD::Vec3<vertexType> cmin(0,0,0);
	bool noStartingVertex = true;
    VHACD::Vec3<vertexType> temp;
	
	if (!sti) return aabbHalfExtents;
				
	const int subParts = sti->getNumSubParts();
	const unsigned char *pverts;int numVerts;PHY_ScalarType type;int stride;
	const unsigned char *pinds;int indsStride;int numTriangles;PHY_ScalarType indsType;	// Unused, but needed
				
	for (int subpart=0;subpart<subParts;subpart++)	{	 
		sti->getLockedReadOnlyVertexIndexBase(&pverts,numVerts,type,stride,&pinds,indsStride,numTriangles,indsType,subpart);
		try	{						
			switch (type)	{
				case PHY_FLOAT:	{
					typedef float glVertex;
					const glVertex* verts = (const glVertex*) pverts; 
					const int vertexDeltaStrideInGLVertexUnits = stride <= sizeof(glVertex)*3 ? 0 : (stride-sizeof(glVertex)*3)/sizeof(glVertex); 			
					
					for (size_t i = 0; i < numVerts;++i)	{						
						temp.X()=(vertexType) *verts++;
						temp.Y()=(vertexType) *verts++;
						temp.Z()=(vertexType) *verts++;						
						verts+=vertexDeltaStrideInGLVertexUnits;
						
						if (noStartingVertex)	{
							noStartingVertex = false;
							cmin = temp;cmax = temp;
						}
						else	{
							if 		(cmax.X()<temp.X()) cmax.X()=temp.X();
							else if (cmin.X()>temp.X())	cmin.X()=temp.X();
							if 		(cmax.Y()<temp.Y()) cmax.Y()=temp.Y();
							else if (cmin.Y()>temp.Y())	cmin.Y()=temp.Y();
							if 		(cmax.Z()<temp.Z()) cmax.Z()=temp.Z();
							else if (cmin.Z()>temp.Z())	cmin.Z()=temp.Z();														
						}
					}					
			 	break;
		 		}
				case PHY_DOUBLE:{
					typedef double glVertex;
					const glVertex* verts = (const glVertex*) pverts; 
					const int vertexDeltaStrideInGLVertexUnits = stride <= sizeof(glVertex)*3 ? 0 : (stride-sizeof(glVertex)*3)/sizeof(glVertex); 			

					for (size_t i = 0; i < numVerts;++i)	{
						temp.X()=(vertexType) *verts++;
						temp.Y()=(vertexType) *verts++;
						temp.Z()=(vertexType) *verts++;						
						verts+=vertexDeltaStrideInGLVertexUnits;
						
						if (noStartingVertex)	{
							noStartingVertex = false;
							cmin = temp;cmax = temp;
						}
						else	{
							if 		(cmax.X()<temp.X()) cmax.X()=temp.X();
							else if (cmin.X()>temp.X())	cmin.X()=temp.X();
							if 		(cmax.Y()<temp.Y()) cmax.Y()=temp.Y();
							else if (cmin.Y()>temp.Y())	cmin.Y()=temp.Y();
							if 		(cmax.Z()<temp.Z()) cmax.Z()=temp.Z();
							else if (cmin.Z()>temp.Z())	cmin.Z()=temp.Z();														
						}
					}
			 	break;
		 		}
				default:
				btAssert((type == PHY_FLOAT) || (type == PHY_DOUBLE));
			}
		}
		catch (...) {
		}
		
		sti->unLockReadOnlyVertexBase(subpart);
	}	
	
	aabbHalfExtents = (cmax - cmin)*vertexType(0.5);
    aabbHalfExtents = VHACD::Vec3<VHACD::Real>(aabbHalfExtents.X()*scalingFactorToApply.X(),aabbHalfExtents.Y()*scalingFactorToApply.Y(),aabbHalfExtents.Z()*scalingFactorToApply.Z());
	if (pOptionalCenterPointOut) {
		*pOptionalCenterPointOut =  (cmax + cmin)*vertexType(0.5);
        *pOptionalCenterPointOut = VHACD::Vec3<VHACD::Real>(pOptionalCenterPointOut->X()*scalingFactorToApply.X(),pOptionalCenterPointOut->Y()*scalingFactorToApply.Y(),pOptionalCenterPointOut->Z()*scalingFactorToApply.Z());
	}	
	return aabbHalfExtents;
}
static btVector3 GetAabbHalfExtents(const std::vector< VHACD::Vec3<VHACD::Real> >& points,btVector3* pOptionalCenterPointOut)	{
    return GetAabbHalfExtents(points.size()==0 ? NULL : (const VHACD::Vec3<VHACD::Real>*)&points[0],points.size(),pOptionalCenterPointOut);
}
static btVector3 GetAabbHalfExtents(const VHACD::Vec3<VHACD::Real>* points,const size_t nPoints,btVector3* pOptionalCenterPointOut)	{
    btVector3 minValue,maxValue,tempVert;
    if (points && nPoints>0)	{
            minValue=maxValue=btVector3(points[0].X(),points[0].Y(),points[0].Z());
    }
    else minValue=maxValue=btVector3(0,0,0);
    for (size_t i=1; i<nPoints; i++)	{
        tempVert = btVector3(points[i].X(),points[i].Y(),points[i].Z());
        minValue.setMin(tempVert);
        maxValue.setMax(tempVert);
    }
    if (pOptionalCenterPointOut) *pOptionalCenterPointOut = (minValue+maxValue)*btScalar(0.5);
    return (maxValue-minValue)*btScalar(0.5);
}

static void ScalePoints(std::vector < VHACD::Vec3<VHACD::Real> >& verts,const VHACD::Vec3<VHACD::Real>& scalingFactor)	{
	for (size_t i=0,sz=verts.size();i<sz;i++)	{
        VHACD::Vec3<VHACD::Real>& v = verts[i];
		v.X()*=scalingFactor.X();
		v.Y()*=scalingFactor.Y();
		v.Z()*=scalingFactor.Z();
	}
}
static void ShiftPoints(std::vector < VHACD::Vec3<VHACD::Real> >& verts,const VHACD::Vec3<VHACD::Real>& shift)	{
	for (size_t i=0,sz=verts.size();i<sz;i++)	{
        VHACD::Vec3<VHACD::Real>& v = verts[i];
		v+=shift;
	}
}

struct Less{
	inline bool operator()(const int& a,const int& b) const {return a<b;}
};

inline static bool AreEqual(const VHACD::Real s1,const VHACD::Real s2,const VHACD::Real eps=SIMD_EPSILON)	{
	return fabs(s1-s2)< eps;
}

inline static bool AreEqual(const VHACD::Vec3<VHACD::Real>& v1,const VHACD::Vec3<VHACD::Real>& v2,const VHACD::Vec3<VHACD::Real>& eps=VHACD::Vec3<VHACD::Real>(SIMD_EPSILON,SIMD_EPSILON,SIMD_EPSILON) )	{
	return (AreEqual(v1.X(),v2.X(),eps.X()) && AreEqual(v1.Y(),v2.Y(),eps.Y()) && AreEqual(v1.Z(),v2.Z(),eps.Z()) );
}

// This same method can be used to decimate the mesh too. Note that if the mesh has texcoords or other additional data, these will be broken after this method is called.
void RemoveDoubleVertices(std::vector< VHACD::Vec3<VHACD::Real> >& verts,std::vector< VHACD::Vec3<long> >& triangles,const bool removeDegenerateTrianglesToo=false,const VHACD::Vec3<VHACD::Real>& eps=VHACD::Vec3<VHACD::Real>(SIMD_EPSILON,SIMD_EPSILON,SIMD_EPSILON))	{
	
	typedef long U;
    typedef VHACD::Vec3<VHACD::Real> Vector3;
		
	U vertsSize = (U) verts.size();
	size_t trianglesSize = triangles.size();
	
	std::vector<U> S(vertsSize,vertsSize);	
	
	U cnt = 0, minDoubleVert = vertsSize, maxSingleVert = 0;
	
	std::vector< Vector3 > newVerts; 							
	newVerts.reserve(vertsSize);

	//---------- Optional Stuff--------------------
	// Basically we want to average the double verts before removing them
	// We can set to false the following line to speed up code...
	const bool useFineTuningForVerts = true;
	const bool useSomeFineTuning = useFineTuningForVerts;
		
	std::vector < unsigned int > C;	// counts the number a given vert is repeated
	if (useSomeFineTuning) C.reserve(vertsSize);

	//----------------------------------------------
	for (U t=0;t<vertsSize;t++)	{
		if (S[t]!=vertsSize) continue;
		S[t]=cnt;
		maxSingleVert = t;
		const Vector3* vertsT = &verts[t];
		if (useSomeFineTuning) C.push_back(1);	
		newVerts.push_back(*vertsT);
		for (U u=t+1;u<vertsSize;u++)	{
			if (AreEqual(*vertsT,verts[u],eps))	{
				S[u] = cnt;
				if (minDoubleVert > u) minDoubleVert = u;
				if (useSomeFineTuning)	{
					C[cnt]+=1;
					if (useFineTuningForVerts) newVerts[cnt]+=verts[u];	//These verts should be equal... however if we pass a bigger "eps" they could differ...
				}
			}
		}
		cnt++;
	} 

	const size_t newVertsSize = cnt;
	if (useSomeFineTuning)	{	// We must average the data so that we can replace multiple equal vertices with their average value
		for (size_t i = 0;i<newVertsSize;i++)	{
			if (C[i]==1) continue;
			const float count = (float) C[i];
			if (useFineTuningForVerts) newVerts[i]/=count;		
		}
	}
		
	
	{
        std::vector< VHACD::Vec3<U> > newTriangles(trianglesSize);
		size_t numDegenerateTriangles = 0;	// Never used
		size_t cnt = 0;
		for (size_t t=0;t<trianglesSize;t++)	{
            const VHACD::Vec3<U>& tri = triangles[t];
            VHACD::Vec3<U>& newTri = newTriangles[cnt];
			newTri.X() = S[tri.X()];
			newTri.Y() = S[tri.Y()];
			newTri.Z() = S[tri.Z()];
			if (removeDegenerateTrianglesToo && (newTri.X()==newTri.Y() || newTri.Y()==newTri.Z() || newTri.X()==newTri.Z())) {
				++numDegenerateTriangles;continue;
			}
			++cnt;
		}
		if (trianglesSize!=cnt) newTriangles.resize(cnt);
		triangles = newTriangles;
	}

	
	verts = newVerts;	

}

static void VHACDCallBackFunction(const char * msg)
{
	printf("%s",msg);
}

static void PerformStanMelaxVolumeIntegration(const std::vector < VHACD::Vec3<VHACD::Real> >& points,const std::vector < VHACD::Vec3<long> >& triangles,btScalar& volumeOut,btVector3* pCentroidOut=NULL)	{
        return PerformStanMelaxVolumeIntegration(points.size()==0? NULL : (const VHACD::Vec3<VHACD::Real>*)&points[0],
                                                 points.size(),
                                                 triangles.size()==0?NULL:(const VHACD::Vec3<long>*)&triangles[0],
                                                 triangles.size(),
                                                 volumeOut,pCentroidOut);
}
static void PerformStanMelaxVolumeIntegration(const VHACD::Vec3<VHACD::Real>* points,const size_t nPoints,const VHACD::Vec3<long>* triangles,const size_t nTriangles,btScalar& volumeOut,btVector3* pCentroidOut=NULL)	{
        const size_t numFaces = nTriangles;
        volumeOut = btScalar(0.);
        btVector3 com(0., 0., 0.);
        btVector3 v0,v1,v2;
        btScalar vol;
        for (size_t j=0; j < numFaces; j++) {
            const VHACD::Vec3<long>& tri(triangles[j]);
            const VHACD::Vec3<VHACD::Real>& p0 = points[tri[0]];
            const VHACD::Vec3<VHACD::Real>& p1 = points[tri[1]];
            const VHACD::Vec3<VHACD::Real>& p2 = points[tri[2]];
            v0.setValue(p0.X(),p0.Y(),p0.Z());
            v1.setValue(p1.X(),p1.Y(),p1.Z());
            v2.setValue(p2.X(),p2.Y(),p2.Z());

            vol = v0.triple(v1,v2);
            volumeOut += vol;
            com += vol * (v0 + v1 +v2);
        }
        com /= volumeOut * btScalar(4.);
        volumeOut /= btScalar(6.);
        if (pCentroidOut) *pCentroidOut=com;
}

public:
//TODO: Better adding strides and VertsType template too... (and accepting array pointers)
template < typename IndsType > inline static void PerformStanMelaxVolumeIntegration(const btAlignedObjectArray < btVector3 >& verts,const btAlignedObjectArray < IndsType >& inds,btScalar& volumeOut,btVector3* pCentroidOut=NULL)	{
		volumeOut = btScalar(0.);
		btVector3 com(0., 0., 0.);
		btScalar vol;
		int j3;
		for (int j=0,sz = inds.size()/3; j < sz; j++) {
			j3 = j*3;	// Division + mult should avoid errors in case inds.size()%3!=0...
			const btVector3& v0(verts[inds[j3]]);
			const btVector3& v1(verts[inds[j3+1]]);
			const btVector3& v2(verts[inds[j3+2]]);
			
			vol = v0.triple(v1,v2);
			volumeOut += vol;
			com += vol * (v0 + v1 +v2);
		}
		com /= volumeOut * btScalar(4.);
		volumeOut /= btScalar(6.);
		if (pCentroidOut) *pCentroidOut=com;
}
static btVector3 GetAabbHalfExtents(const btAlignedObjectArray< btVector3 >& points,btVector3* pOptionalCenterPointOut)	{
 	btVector3 minValue,maxValue;
	const int nPoints(points.size());
	if (nPoints>0)	{
			minValue=maxValue=points[0];
	}
	else minValue=maxValue=btVector3(0,0,0);
 	for (size_t i=1; i<nPoints; i++)	{
 		const btVector3& tempVert(points[i]);
		minValue.setMin(tempVert);
		maxValue.setMax(tempVert);
	}
	if (pOptionalCenterPointOut) *pOptionalCenterPointOut = (minValue+maxValue)*btScalar(0.5);			
	return (maxValue-minValue)*btScalar(0.5);
}
};
#pragma endregion

//Reference Usage:
/*
    std::vector< VHACD::Vec3<VHACD::Real> > points;
    std::vector< VHACD::Vec3<long> > triangles;

    LoadOFF(fileName, points, triangles, invert);
    //SaveOFF(outOFFFileName.c_str(), points, triangles);


    bool decimatedMeshComputed = false;
    size_t nTriangles = triangles.size();
    size_t nPoints    = points.size();
    VHACD::Vec3<VHACD::Real> * pPoints    = &points[0];
    VHACD::Vec3<long> *        pTriangles = &triangles[0];

    if (targetNTrianglesDecimatedMesh > 0 && targetNTrianglesDecimatedMesh < nTriangles)
    {
        decimatedMeshComputed = true;

        VHACD::MeshDecimator myMDecimator;
        myMDecimator.SetCallBack(&CallBack);
        myMDecimator.Initialize(nPoints, nTriangles, &points[0], &triangles[0]);
        myMDecimator.Decimate(0, targetNTrianglesDecimatedMesh);

        nTriangles = myMDecimator.GetNTriangles();
        nPoints    = myMDecimator.GetNVertices();
        pPoints     = new VHACD::Vec3<VHACD::Real>[nPoints];
        pTriangles  = new VHACD::Vec3<long>[nTriangles];
        myMDecimator.GetMeshData(pPoints, pTriangles);
    }

    VHACD::Mesh mesh;
    mesh.ResizePoints(nPoints);
    mesh.ResizeTriangles(nTriangles);
    for(size_t p = 0; p < nPoints;    ++p) mesh.SetPoint(p, pPoints[p]);
    for(size_t t = 0; t < nTriangles; ++t) mesh.SetTriangle(t, pTriangles[t]);
    if (decimatedMeshComputed)
    {
        delete [] pPoints;
        delete [] pTriangles;
    }

    mesh.SaveOFF(outOFFFileNameDecimated);

    std::vector< VHACD::Mesh * > parts;
    VHACD::ApproximateConvexDecomposition(mesh, depth, posSampling, angleSampling, posRefine, angleRefine, alpha, concavity, parts, &CallBack);

    const size_t nParts = parts.size();

    std::string outFileName = folder + PATH_SEP + file.substr(0, file.find_last_of(".")) + "_VHACD.wrl";
    std::string outFileNameCH = folder + PATH_SEP + file.substr(0, file.find_last_of(".")) + "_VHACD_CHs.wrl";
    std::ofstream fout(outFileName.c_str());
    std::ofstream foutCH(outFileNameCH.c_str());
    if (fout.is_open() && foutCH.is_open())
    {
        VHACD::Material mat;
        for(size_t p = 0; p < nParts; ++p)
        {
            mat.m_diffuseColor.X() = mat.m_diffuseColor.Y() = mat.m_diffuseColor.Z() = 0.0;
            while (mat.m_diffuseColor.X() == mat.m_diffuseColor.Y() ||
                   mat.m_diffuseColor.Z() == mat.m_diffuseColor.Y() ||
                   mat.m_diffuseColor.Z() == mat.m_diffuseColor.X()  )
            {
                mat.m_diffuseColor.X() = (rand()%100) / 100.0;
                mat.m_diffuseColor.Y() = (rand()%100) / 100.0;
                mat.m_diffuseColor.Z() = (rand()%100) / 100.0;
            }
            parts[p]->SaveVRML2(fout, mat);
            VHACD::Mesh ch;
            parts[p]->ComputeConvexHull(ch);
            ch.SaveVRML2(foutCH, mat);
        }
        fout.close();
        foutCH.close();
    }
    for(size_t p = 0; p < nParts; ++p)
    {
        delete parts[p];
    }
*/



#endif //BTVHACDCOMPOUNDSHAPE_H__



